;; -*- Mode: Emacs-Lisp; Coding: iso-2022-7bit -*-

;; $B@_Dj$7$J$$$HF0$-$^$;$s(B & default$B%U%!%$%k$OI,?\(B
(defvar auto-complete-template-dir nil) 
;; $B@_Dj$7$?J}$,$$$$$d$D(B
(defvar auto-complete-key-file-alist nil)

;; $B%+%9%?%^%$%68~$1(B (key$B$O0lJ8;z$G$"$k$3$H(B)
(defvar auto-complete-key-begin "")
(defvar auto-complete-key-end "")
(defvar auto-complete-key-position "")
(defvar auto-complete-key-noprefix "")
(defvar auto-complete-macro-inherit-file-directive "inherit")
(defvar auto-complete-dynamic-macro-eval "eval")
(defvar auto-complete-dynamic-macro-eval-noprefix "eval-noprefix")


;; $BBeF~$,2DG=$J%3%s%H%m!<%kJ8;z(B ($B%3!<%I$N=gHV$KE83+$5$l$F$f$/(B)
(defvar auto-complete-key-start-code 16)              ; $B@)8fJ8;z$N3+;O%3!<%I(B(10$B?JK!(B) Default-> 16(^P)
(defvar auto-complete-key-code-count 8)               ; $B@_Dj$G$-$k8D?t(B Default$B$O(B->8 ($B@_Dj2DG=J8;z(B ^P ^Q ^R ^S ^T ^U ^V ^W) $B"((B $B$?$/$5$s@_Dj$9$k$HCY$/$J$k$H;W$$$^$9(B


;; ----- Use Variable
(defvar auto-complete-key-collection nil) ; keywordtree ...  list of hash (major-mode, keywords' list's)


;; --------------------------------------------------------------------
;; Edit Mode for Auto Complete
;; --------------------------------------------------------------------


(defvar auto-complete-edit-mode-syntax-table nil
  "syntax table in use in mynote-mode buffers.")

(if auto-complete-edit-mode-syntax-table
    ()
  (setq auto-complete-edit-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?_ "w" mynote-card-mode-syntax-table))


;;-----
;;faces
;;-----


(defface auto-complete-edit-begin-end-face
  '((t (:background "yellow" :bold t)))
  "Used for displaying the region.")
(defvar auto-complete-edit-begin-end-face 'auto-complete-edit-begin-end-face)


(defface auto-complete-edit-separator-face
  '((t (:background "gainsboro")))
  "Used for separeted several templates.")
(defvar auto-complete-edit-separator-face 'auto-complete-edit-separator-face)

(defface auto-complete-edit-key-face
  '((t (:background "sky blue")))
  "Used for displaying the region.")
(defvar auto-complete-edit-key-face 'auto-complete-edit-key-face)

(defface auto-complete-edit-position-face
  '((t (:background "light pink")))
  "Used for displaying the region.")
(defvar auto-complete-edit-position-face 'auto-complete-edit-position-face)

(defvar auto-complete-edit-font-lock-keywords
  (list
   (cons (concat "["
                 (let (words)
                   (dotimes (i auto-complete-key-code-count)
                     (setq words (concat (format "%c" (+ auto-complete-key-start-code i)) words)))
                   words)
                 "]") auto-complete-edit-key-face)
   (cons (concat "[" auto-complete-key-begin auto-complete-key-end "]")
                 auto-complete-edit-begin-end-face)
   (cons "^-\\*.*$" auto-complete-edit-separator-face)
   (cons auto-complete-key-position
                 auto-complete-edit-position-face)))

(defun auto-complete-edit-mode ()
  (interactive)
  ; Initializing
  (kill-all-local-variables)
  ; Setting up font-locking
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(auto-complete-edit-font-lock-keywords nil t nil nil))

  ; Setting up syntax
  (set-syntax-table auto-complete-edit-mode-syntax-table)
  ; Final stuff, then we're done
  (setq mode-name "AComplete-Edit")
  (setq major-mode 'auto-complete-edit-mode)
  (make-local-variable 'font-lock-defaults))

;; --------------------------------------------------------------------
;; Exec Auto Complete
;; --------------------------------------------------------------------


;; ---- Collect Keyword


(defun auto-complete ()
  (interactive)
  (let ((key-alist (auto-complete-get-keys-from-mode major-mode))
        key
        template)
    (setq key (completing-read "Key : " key-alist nil t))
    (setq template (auto-complete-get-template-from-key key (cdr (assoc key key-alist))))
    (if (null template)
        (message "can't find template of key '%s'" key)
      (auto-complete-insert-from-template template))))


(defun auto-complete-open-template ()
  (interactive)
  (let ((file (cdr (assoc major-mode auto-complete-key-file-alist))))
    (if (null file)
        (setq file "default"))
    (setq file (expand-file-name file auto-complete-template-dir))
    (split-window-vertically)
    (other-window 1)
    (find-file file)
    (message file)))

(defun auto-complete-clear ()
  (interactive)
  (setq auto-complete-key-collection nil))

(defun auto-complete-get-keys-from-mode (mode)
  (let ((key-alist (cdr (assoc mode auto-complete-key-collection))))
    (if (null key-alist)
        (let ((file (cdr (assoc mode auto-complete-key-file-alist))))
          (if (null file)
              (setq file "default"))
          (setq key-alist (auto-complete-collect-key-from-file file))
          (setq auto-complete-key-collection
                (cons (cons mode key-alist)
                      auto-complete-key-collection ))))
    key-alist))

(defun auto-complete-collect-key-from-file(filename &optional keys)
  (let (key
        inherit-file-list)
    (setq filename (expand-file-name filename auto-complete-template-dir))
    (with-temp-buffer
      (insert-file-contents filename)
      ;; collect keywords
      (goto-char (point-min))
      (while (re-search-forward (concat "^" auto-complete-key-begin
                                        "\\([^ \t\n]+\\)[ \t]*$")
                              nil t)
        (setq key (match-string 1))
        (if (not (assoc key keys))
            (setq keys (cons (cons (match-string 1) filename) keys))))
      ;; collect inherit
      (goto-char (point-min))
      (let ((beg (point))
            (end (if (search-forward auto-complete-key-begin nil t)
                     (progn
                       (beginning-of-line)
                       (point))
                   (goto-char (point-max))
                   (point))))
        (save-excursion
          (save-restriction
            (narrow-to-region beg end)
            (goto-char (point-min))
            (while (re-search-forward (concat
                                       "^"
                                       auto-complete-macro-inherit-file-directive
                                       "[ \t]*\\([^ \t\n]+\\)")
                                      nil t)
              (setq inherit-file-list (cons (match-string 1) inherit-file-list))))))
      (if inherit-file-list
          (while inherit-file-list
            (setq keys (auto-complete-collect-key-from-file
                        (car inherit-file-list)
                        keys))
            (setq inherit-file-list (cdr inherit-file-list)))))
      keys))

(defun auto-complete-get-template-from-key (key filename)
  ;; get contents of key from file (if can't find it search for inherit file defined in file)
  ;; inherit memo --> howm
  (with-temp-buffer
    (insert-file-contents filename)
    (goto-char (point-min))
    (if (re-search-forward (concat "^" auto-complete-key-begin key "[ \t]*$")
                           nil t)
        (progn
          (forward-line 1)
          (beginning-of-line)
          (let ((beg (point))
                (end (if (re-search-forward auto-complete-key-end  nil t)
                         (progn
                           (forward-char (- (length auto-complete-key-end)))
                           (point)))))
            (if end
                (save-excursion
                  (save-restriction
                    (narrow-to-region beg end)
                    (buffer-string)))
              nil)))
      nil)))



(defun auto-complete-insert-from-template (template)
  "insert template , and if defined ,jump to point"
  (dotimes (i auto-complete-key-code-count)
    (setq template (auto-complete-insert-from-template-eval-macro
                    template
                    (format "%c" (+ auto-complete-key-start-code i)))))
  (let ((lines (progn
                 (if (string-match "\n$" template)
                     (setq template (concat template "\n")))
                 (if (string-match "^\n" template)
                     (setq template (concat "\n" template)))
                 (split-string template "\n")))
        line
        (count 0)
        no-prefix
        no-prefix-pos
        (pos (string-match auto-complete-key-position template))
        (prefix (let ((end (point))
                      (beg (save-excursion
                             (beginning-of-line) (point))))
                  (if (< beg end)
                      (buffer-substring beg end)))))
    (while lines
      (setq line (car lines))
      (if (setq no-prefix-pos (string-match auto-complete-key-noprefix line))
          (progn
            (setq no-prefix (not no-prefix))
            (setq line(replace-match "" t t line))))
      (and prefix
           (not (eq count 0))
           (not no-prefix)
           (not no-prefix-pos)
           (insert prefix))
      (insert line)
      (setq lines (cdr lines))
      (setq count (1+ count))
      (if lines (insert "\n")))
    (if pos
        ;; jump to  position and delete it
        (progn
          (re-search-backward auto-complete-key-position nil t)
          (delete-char (length auto-complete-key-position))))))

(defun auto-complete-insert-from-template-eval-macro (template ctl-char)
  (let ((pos-1st-beg (string-match ctl-char template))
        pos-1st-end
        body)
    ;; --- init body
    (if pos-1st-beg
        (let ((delimiter (substring template (+ pos-1st-beg 1) (+ pos-1st-beg 2)))
              (pos pos-1st-beg)
              title
              default)
          (dotimes (i 3)
            (setq pos (string-match delimiter template (1+ pos)))
            (if (null pos) (error (concat "Can't find next '" delimiter "' in template '" template  "'"))))
          (setq pos-1st-end (1+ pos))
          (let ((array (split-string (substring template pos-1st-beg pos-1st-end) delimiter)))
            (setq title (car (cdr array)))
            (setq default (car (cddr array))))
          ;; define body
          (setq body (auto-complete-insert-from-template-define-body title default))))
    ;; --- re-init template
    (if body
        (progn
          (setq template (concat (substring template 0 pos-1st-beg)
                                 body
                                 (substring template pos-1st-end)))
          (while (string-match ctl-char template)
            (setq template (replace-match body t t template)))))
    template))

(defun auto-complete-insert-from-template-define-body (title default)
  (let (value)
    (cond ((string= title auto-complete-dynamic-macro-eval)
           (setq value (eval (read default))))
          ((string= title auto-complete-dynamic-macro-eval-noprefix)
           (setq value (eval (read default)))
           (if value
               (setq value (concat auto-complete-key-noprefix value auto-complete-key-noprefix))))
          (t
           (setq value (read-string (concat title " : ") default))))
    (if (null value)
        ""
      value)))


(defun auto-complete-sample-get-c-header-upper ()
   (let ((file (file-name-nondirectory (buffer-file-name))))
     (while (string-match "\\." file)
       (setq file (replace-match "_" t t file)))
     (upcase file)))



(provide 'auto-complete)
